package ch.unige.solidify.util;

import java.io.BufferedOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.boot.web.client.RestTemplateCustomizer;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.client.ClientHttpRequestInterceptor;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.stereotype.Component;
import org.springframework.web.client.RequestCallback;
import org.springframework.web.client.ResponseExtractor;
import org.springframework.web.client.RestTemplate;

import com.nimbusds.jwt.SignedJWT;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.interceptor.BearerAuthorizationInterceptor;
import ch.unige.solidify.interceptor.HeaderRequestInterceptor;
import ch.unige.solidify.interceptor.PlusEncoderInterceptor;
import ch.unige.solidify.model.sso.SsoUser;

@Component
public class RestClientTool {

  private static class UTF8MessageConverterCustomizer implements RestTemplateCustomizer {
    @Override
    public void customize(RestTemplate restTemplate) {
      final List<HttpMessageConverter<?>> converters = restTemplate.getMessageConverters();
      converters.removeIf(httpMessageConverter -> httpMessageConverter instanceof StringHttpMessageConverter);
      converters.add(0, new StringHttpMessageConverter(StandardCharsets.UTF_8));
    }
  }

  private static final Logger log = LoggerFactory.getLogger(RestClientTool.class);

  @Value("${solidify.oauth2.accesstoken:#{null}}")
  private String accessToken;

  private String trustedToken;

  public RestClientTool() {
    if (this.accessToken == null) {
      this.accessToken = System.getProperty("solidify.oauth2.accesstoken");
    }
  }

  public void setTrustedToken(String trustedToken) {
    this.trustedToken = trustedToken;
  }

  public boolean hasTrustedToken() {
    return !StringTool.isNullOrEmpty(this.trustedToken);
  }

  public void downloadContentWithToken(URI uri, Path path) {
    ResponseExtractor<Void> responseExtractor = this.buildExtractor(uri, path);
    final RestTemplate restTemplate =  this.getRestTemplateBuilder(uri.toString()).build();
    this.launchDownload(uri, responseExtractor, restTemplate);
  }

  public void downloadContentWithoutToken(URI uri, Path path) {
    ResponseExtractor<Void> responseExtractor = this.buildExtractor(uri, path);
    final RestTemplate restTemplate =  new RestTemplateBuilder().build();
    this.launchDownload(uri, responseExtractor, restTemplate);
  }

  private void launchDownload(URI uri, ResponseExtractor<Void> responseExtractor, RestTemplate restTemplate) {
    // Optional Accept header
    final RequestCallback requestCallback = request -> request.getHeaders()
            .setAccept(Arrays.asList(MediaType.APPLICATION_OCTET_STREAM, MediaType.ALL));
    restTemplate.execute(uri, HttpMethod.GET, requestCallback, responseExtractor);
  }

  private ResponseExtractor<Void> buildExtractor(URI uri, Path path) {
    return response -> {
      final byte[] data = new byte[SolidifyConstants.BUFFER_SIZE];
      try (final InputStream inputStream = response.getBody();
           final FileOutputStream fileOutputStream = new FileOutputStream(path.toFile());
           final BufferedOutputStream bufferedOutputStream = new BufferedOutputStream(fileOutputStream, SolidifyConstants.BUFFER_SIZE)) {
        int nRead;
        while ((nRead = inputStream.read(data, 0, data.length)) != -1) {
          bufferedOutputStream.write(data, 0, nRead);
        }
      } catch (final IOException e) {
        throw new SolidifyRuntimeException("Error downloading " + uri.toString() + " to " + path.toString() + uri, e);
      }
      return null;
    };
  }

  public RestTemplate getClient(String url) {
    return this.getRestTemplateBuilder(url).build();
  }

  public RestTemplate getClientWithPlusEncoderInterceptor(String url) {
    return this.getRestTemplateBuilder(url).additionalInterceptors(new PlusEncoderInterceptor()).build();
  }

  private RestTemplateBuilder getRestTemplateBuilder(String urlString) {
    RestTemplateBuilder restTemplateBuilder = new RestTemplateBuilder();
    restTemplateBuilder = this.addToken(urlString, restTemplateBuilder);
    restTemplateBuilder = this.addUTF8Converter(restTemplateBuilder);
    return restTemplateBuilder;
  }

  private RestTemplateBuilder addUTF8Converter(RestTemplateBuilder restTemplateBuilder) {
    return restTemplateBuilder.additionalCustomizers(new UTF8MessageConverterCustomizer());
  }

  private RestTemplateBuilder addToken(String urlString, RestTemplateBuilder restTemplateBuilder) {
    if (!StringTool.isNullOrEmpty(this.accessToken)) {
      log.debug("Use shared token: {}", this.parseToken(this.accessToken));
      // OAuth2 without end user: shared AccesToken
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(new BearerAuthorizationInterceptor(this.accessToken));
    } else if (this.trustedToken != null) {
      log.debug("Use trusted user token: {}", this.parseToken(this.trustedToken));
      // DLCM : Access token of trusted user
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString).additionalInterceptors(
              new BearerAuthorizationInterceptor(this.trustedToken));
    } else if (UserTool.getUserPrincipal() instanceof SsoUser) {
      log.debug("Use OIDC user {}", UserTool.getUserPrincipal().getUniqueId());
      // Rock Solid : AccesToken, usercontext, useruniqueID of end user for OIDC Autho2 protected MS
      final SsoUser user = (SsoUser) UserTool.getUserPrincipal();
      final List<ClientHttpRequestInterceptor> intercepters = new ArrayList<>();
      if (user.getUniqueId() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_UNIQUE_ID,
                user.getUniqueId()));
      }
      if (user.getUserContext() != null) {
        intercepters.add(new HeaderRequestInterceptor(SsoUser.CONNECTED_USER_CONTEXT_SOURCE_NAME,
                user.getUserContext()));
      }
      if (UserTool.getUserPrincipal().getAccessToken() != null) {
        intercepters.add(new BearerAuthorizationInterceptor(
                UserTool.getUserPrincipal().getAccessToken()));
      }
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString)
              .additionalInterceptors(intercepters);
    } else {
      log.debug("Use no authorization");
      // No Auth
      restTemplateBuilder = restTemplateBuilder.rootUri(urlString);
    }
    return restTemplateBuilder;
  }

  private String parseToken(String token) {
    try {
      return SignedJWT.parse(token).getJWTClaimsSet().toString();
    } catch (ParseException e) {
      return e.getMessage();
    }
  }

  private String getUrl(URL url) {
    String rootUrl = url.getProtocol() + "://" + url.getHost();
    if (url.getPort() > 0) {
      rootUrl += ":" + url.getPort();
    }
    return rootUrl + url.getPath();
  }

}
